@@ -1,3 +1,4 @@
+from operator import pos
 import os
 from miditoolkit.midi import parser
 from miditoolkit.midi import containers as ct
@@ -5,6 +6,9 @@ from collections import defaultdict
 import numpy as np
 from copy import deepcopy
 
+from numpy.core.defchararray import ljust
+from numpy.lib.arraysetops import isin
+
 from deepnote.const import Constants
 from deepnote.modules import *
 from deepnote import utils
@@ -32,23 +36,22 @@ class MusicRepr:
                 if len(bars) == len(other_bars):
                     for bar1, bar2 in zip(bars, other_bars):
                         if not utils.compare_bars(bar1, bar2):
-                            print(bar1, bar2)
                             return False
                     return True
         return False
 
     def find_beat_index(self, beat):
         query = beat * self.const.unit
-        prev_bar = -self.const.n_bar_steps
-        t = 0
+        n_bars = 0
+        prev_idx = 0
         for i, e in enumerate(self.events):
-            if isinstance(e, Metric) and e.position == 0:
-                prev_bar += self.const.n_bar_steps
-                t = prev_bar
-            if isinstance(e, Metric) and e.position > 0:
-                t = prev_bar + e.position
-            if query < t:
-                return i - 1
+            if isinstance(e, Metric):
+                prev_pos = e.position
+                if e.position == 0:
+                    n_bars += 1
+                if query < (n_bars-1)*self.const.n_bar_steps + prev_pos:
+                    return prev_idx
+                prev_idx = i
         return -1
 
     def slice_by_index(self, start, end):
@@ -163,16 +166,14 @@ class MusicRepr:
                 marker.text = marker.text[6:]
                 events[marker.time // const.step] += [marker]
 
+        for pos in range(0, sorted(events)[-1] + 1, const.n_bar_steps):
+            if pos not in events:
+                events[pos] = []
+
         res = []
-        pos_indices = sorted(list(events.keys()))
-        bar_idx = -1
-        for pos_idx in pos_indices:
-            while pos_idx // const.n_bar_steps > bar_idx:
-                res += [Metric()]  ## bar
-                bar_idx += 1
-            
+        for pos_idx in sorted(events):
             notes = []
-            beat = Metric(position=(pos_idx % const.n_bar_steps) + 1)
+            beat = Metric(position=pos_idx % const.n_bar_steps)
             for e in events[pos_idx]:
                 if isinstance(e, Note):
                     notes += [e]
@@ -194,10 +195,9 @@ class MusicRepr:
         for c in cp:
             if c[0] == 0: ## metric
                 if c[1] == 0:
-                    bars += [[Metric()]]
+                    bars += [[Metric.from_cp(c, const=const)]]
                 else:
                     bars[-1] += [Metric.from_cp(c, const=const)]
-            
             elif c[0] == 1:  ## note
                 bars[-1] += [Note.from_cp(c, const=const)]
         res = []
@@ -211,52 +211,49 @@ class MusicRepr:
             const = Constants()
 
         tokens = text.split()
-        n = len(tokens)
-        i = 0
         bars = []
-        while i < n:
-            if tokens[i] == 'Bar':
+        for i, token in enumerate(tokens):
+            if token == 'Bar':
                 bars += [[Metric()]]
 
-            elif tokens[i].startswith('BeatPosition'):
+            elif token.startswith('BeatPosition'):
                 bars[-1] += [Metric(position=int(tokens[i].split('_')[1]))]
 
-            elif tokens[i].startswith('BeatTempo'):
-                assert isinstance(bars[-1][-1], Metric) and bars[-1][-1].position > 0
+            elif token.startswith('BeatTempo'):
+                assert isinstance(bars[-1][-1], Metric)
                 tempo = int(tokens[i].split('_')[1])
                 assert tempo in const.tempo_bins
                 bars[-1][-1].tempo = tempo
 
-            elif tokens[i].startswith('BeatChord'):
-                assert isinstance(bars[-1][-1], Metric) and bars[-1][-1].position > 0
-                chord = tokens[i][9:]
+            elif token.startswith('BeatChord'):
+                assert isinstance(bars[-1][-1], Metric)
+                chord = tokens[i][10:]
                 assert chord in const.chords
                 bars[-1][-1].chord = chord
 
-            elif tokens[i].startswith('NoteInstFamily'):
-                assert tokens[i+1].startswith('NotePitch') and \
-                    tokens[i+2].startswith('NoteDuration') and \
-                        tokens[i+3].startswith('NoteVelocity')
-                
-                inst_family = tokens[i].split('_')[1]
+            elif token.startswith('NoteInstFamily'):
+                inst_family = token.split('_')[1]
                 assert inst_family in const.instruments
-                pitch = int(tokens[i+1].split('_')[1])
+                bars[-1] += [Note(inst_family=inst_family)]
+
+            elif token.startswith('NotePitch'):
+                assert isinstance(bars[-1][-1], Note)
+                pitch = int(token.split('_')[1])
                 assert 0 <= pitch < 128
-                duration = int(tokens[i+2].split('_')[1])
+                bars[-1][-1].pitch = pitch
+
+            elif token.startswith('NoteDuration'):
+                assert isinstance(bars[-1][-1], Note)
+                duration = int(token.split('_')[1])
                 assert duration in const.duration_bins
-                velocity = int(tokens[i+3].split('_')[1])
+                bars[-1][-1].duration = duration
+            
+            elif token.startswith('NoteVelocity'):
+                assert isinstance(bars[-1][-1], Note)
+                velocity = int(token.split('_')[1])
                 assert velocity in const.velocity_bins
+                bars[-1][-1].velocity = velocity
 
-                bars[-1] += [
-                    Note(
-                        inst_family=inst_family,
-                        pitch=pitch,
-                        duration=duration,
-                        velocity=velocity
-                    )
-                ]
-                i += 3
-            i += 1
         res = []
         for bar in bars:
             res += utils.sort_bar_beats(bar)
@@ -269,6 +266,45 @@ class MusicRepr:
             
         return MusicRepr.from_string(' '.join([const.all_tokens[idx] for idx in indices]))
 
+    @staticmethod
+    def from_single_pianoroll(pianoroll : np.array, inst : str, const : Constants = None):
+        def from_pianoroll_bar(bar : np.array):
+            binarized = bar > 0
+            padded = np.pad(binarized, ((0, 0), (1, 1)))
+            diff = np.diff(padded.astype(np.int8), axis=1)
+
+            positives = np.nonzero((diff > 0))
+            pitches = positives[0]
+            note_ons = positives[1]
+            note_offs = np.nonzero((diff < 0))[1]
+
+            poses = {}
+            for idx, pitch in enumerate(pitches):
+                on = note_ons[idx] 
+                off = note_offs[idx]
+                velocity = bar[on, pitch]
+
+                if on not in poses:
+                    poses[on] = []
+                poses[on] += [Note(inst_family=inst, pitch=pitch, duration=off-on, velocity=velocity)]
+
+            events = [Metric()]
+            for pos in sorted(poses):
+                if pos > 0:
+                    events += [Metric(position=pos)]
+                for note in poses[pos]:
+                    events += [note]
+            return events
+
+
+        if const is None:
+            const = Constants()
+        events = []
+        for i in range(0, pianoroll.shape[1], const.n_bar_steps):
+            events += from_pianoroll_bar(pianoroll[:, i:const.n_bar_steps])
+        return MusicRepr(events, const)
+
+
     @staticmethod
     def merge_tracks(tracks : dict, key_inst: str = None):
         insts = list(tracks.keys())
@@ -297,23 +333,47 @@ class MusicRepr:
 
 
     ## Output
-    def to_remi(self, ret='token'):
+    def to_remi(self, ret='token', add_eos=False):
         assert ret in ['token', 'index', 'event']
         res = []
         for e in self.events:
             res += e.to_remi()
+        if add_eos:
+            res += [Event('EOS')]
         if ret != 'event':
             res = [r.to_token() for r in res]
         if ret == 'index':
             res = [self.const.all_tokens.index(tok) for tok in res]
         return res
 
-    def to_cp(self):
+    def to_cp(self, add_eos=False):
         res = []
         for e in self.events:
             res += [e.to_cp(const=self.const)]
+        if add_eos:
+            res += [[2] + [0]*7]
         return np.array(res)
 
+    def to_pianoroll(self, separate_tracks=True, binarize=False):
+        def to_single_pianoroll(track : MusicRepr):
+            roll = np.zeros(shape=(128, track.get_bar_count()*track.const.n_bar_steps))
+            prev_pos = 0
+            prev_bar = -1
+            for e in track.events:
+                if isinstance(e, Metric):
+                    if e.position == 0:
+                        prev_bar += 1
+                    prev_pos = e.position
+                else:
+                    offset = prev_bar*track.const.n_bar_steps + prev_pos
+                    roll[e.pitch, offset:offset+e.duration] = 1 if binarize else e.velocity
+            return roll
+
+        if separate_tracks:
+            tracks = self.separate_tracks()
+            return dict([(inst, to_single_pianoroll(tracks[inst])) for inst in tracks])
+        return to_single_pianoroll(self)
+        
     def to_midi(self, output_path=None):
         midi = parser.MidiFile()
 
@@ -326,13 +386,14 @@ class MusicRepr:
         prev_pos = 0
         for e in self.events:
             if isinstance(e, Metric):
+                prev_pos = e.position
                 if e.position == 0:
                     n_bars += 1
                 if e.tempo:
-                    tempos += [ct.TempoChange(e.tempo, e.position*self.const.step + (n_bars-1)*self.const.bar_resol)]
+                    tempos += [ct.TempoChange(e.tempo, prev_pos*self.const.step + (n_bars-1)*self.const.bar_resol)]
                 if e.chord:
-                    chords += [ct.Marker('Chord_'+e.chord, e.position*self.const.step + (n_bars-1)*self.const.bar_resol)]
-                prev_pos = e.position
+                    chords += [ct.Marker('Chord_'+e.chord, prev_pos*self.const.step + (n_bars-1)*self.const.bar_resol)]
+                
             
             if isinstance(e, Note):
                 s = self.const.step * prev_pos + (n_bars-1)*self.const.bar_resol
@@ -345,8 +406,11 @@ class MusicRepr:
                 
         tempos.sort(key=lambda x: x.time)
         chords.sort(key=lambda x: x.time)
+        max_tick = max(
+            chords[-1].time if len(chords) > 0 else 0, 
+            tempos[-1].time if len(tempos) > 0 else 0
+        )
         
-        max_tick = 0
         instruments = []
         for k, v in instr_notes.items():
             k = self.const.instruments.index(k)
